"""
@Author = 'Michael Stanley'

============ Change Log ============
01/17/2023 = Created

============ License ============
Copyright (C) 2023 Michael Stanley

This file is part of wmul_click_utils.

wmul_click_utils is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_click_utils is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_click_utils. If not, see <https://www.gnu.org/licenses/>. 
"""
import click
import pytest
from click.testing import CliRunner
from wmul_click_utils import DATE
from wmul_test_utils import make_namedtuple


@pytest.fixture(scope="function")
def setup_date():
    @click.command()
    @click.option("--foo", type=DATE, help="foo help")
    def cli(foo):
        click.echo(f"foo={foo.strftime('%Y-%m-%d')}")

    runner = CliRunner()

    return make_namedtuple(
        "setup_date",
        cli=cli,
        runner=runner
    )


def test_good_date(setup_date):
    result = setup_date.runner.invoke(
        setup_date.cli, 
        ["--foo", "2023-01-17"]
    )

    assert not result.exception
    assert "foo=2023-01-17" in result.output

def test_incomplete_date(setup_date):
    result = setup_date.runner.invoke(
        setup_date.cli, 
        ["--foo", "2023-01"]
    )

    assert result.exception
    assert "2023-01 is not a valid Date Type. It must be in the format "\
        "YYYY-MM-DD." in result.output


def test_gibberish_date(setup_date):
    result = setup_date.runner.invoke(
        setup_date.cli, 
        ["--foo", "Earth"]
    )

    assert result.exception
    assert "Earth is not a valid Date Type. It must be in the format "\
        "YYYY-MM-DD." in result.output



"""
@Author = 'Michael Stanley'

These are utilities for use with Python's Click cli library.

MXWith based upon https://stackoverflow.com/a/44349292/521402 
and https://stackoverflow.com/a/51235564/521402

============ Change Log ============
01/18/2023 = Added RequiredUnless

01/17/2023 = Extracted to separate package.

============ License ============
Copyright (C) 2020-2023 Michael Stanley

This file is part of wmul_click_utils.

wmul_click_utils is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_click_utils is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_click_utils. If not, see <https://www.gnu.org/licenses/>. 
"""
__version__ = "0.1.0"
import click
from datetime import datetime

class RequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.required_if = kwargs.pop('required_if')
        assert self.required_if, "'required_if' parameter required"
        kwargs['help'] = f"{kwargs.get('help', '')} NOTE: This argument is " \
                         f"required if {self.required_if} is supplied.".strip()
        super(RequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.required_if in opts

        if other_present:
            if not we_are_present:
                raise click.UsageError(
                    f"Illegal usage: {self.name} is required when "
                    f"{self.required_if} is supplied."
                )
            else:
                self.prompt = None

        return super(RequiredIf, self).handle_parse_result(ctx, opts, args)


class RequiredUnless(click.Option):
    def __init__(self, *args, **kwargs):
        self.required_unless = set(kwargs.pop('required_unless'))
        assert self.required_unless, "'required_unless' parameter required"

        kwargs_help = kwargs.get('help', '')
        req_unless_keys = ", ".join(self.required_unless)
        req_unless_help = " NOTE: This option is required unless "\
                          f"one of {req_unless_keys} is supplied."
        kwargs['help'] = f"{kwargs_help}{req_unless_help}".strip()

        super(RequiredUnless, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt = self.name in opts

        if not current_opt:
            if self.required_unless.isdisjoint(opts):
                raise click.UsageError(
                    f"Illegal usage: '{self.name}' is required unless "
                    f"one of '{', '.join(self.required_unless)}' is "
                        "supplied."
                )
            else:
                self.prompt = None

        return super(RequiredUnless, self).handle_parse_result(ctx, opts, args)


class MXWith(click.Option):
    def __init__(self, *args, **kwargs):
        self.mx_with = kwargs.pop('mx_with')
        assert self.mx_with, "'mx_with' parameter required"

        kwargs_help = kwargs.get('help', '')
        other_mx_keys = ", ".join(self.mx_with)
        mx_help = f" NOTE: This option is mutually exclusive with {other_mx_keys}"
        kwargs['help'] = f"{kwargs_help}{mx_help}".strip()

        super(MXWith, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt = self.name in opts

        for mx_opt in self.mx_with:
            if mx_opt in opts:
                if current_opt:
                    raise click.UsageError(
                        f"Illegal usage: '{self.name}' is mutually exclusive "
                        f"with '{mx_opt}'." 
                    )
                else:
                    self.prompt = None

        return super(MXWith, self).handle_parse_result(ctx, opts, args)


class PairedParamType(click.ParamType):
    name = "Paired Type"

    def convert(self, value, param, ctx):
        value_casefold = value.casefold()
        value_list = value_casefold.split()
        number_of_values = len(value_list)
        if number_of_values % 2 == 0:
            return value_list
        else:
            self.fail(
                f"{value} is not a valid Paired Type. The number of "
                 "space-separated values must be even. E.G. '.wav .mp3' and "
                 "'.wav .mp3 .doc .docx' are valid, while '.wav .mp3 .ogg' is "
                 "not.", 
                 param, 
                 ctx
            )

PAIRED = PairedParamType()


class DateParamType(click.ParamType):
    name = "Date Type"

    def convert(self, value, param, ctx):
        try:
            return datetime.strptime(value, "%Y-%m-%d").date()
        except ValueError:
            self.fail(
                f"{value} is not a valid Date Type. It must be in the format "\
                "YYYY-MM-DD.",
                param,
                ctx
            )

DATE = DateParamType()


"""
@Author = 'Michael Stanley'

============ Change Log ============
01/17/2023 = Created

============ License ============
Copyright (C) 2023 Michael Stanley

This file is part of wmul_click_utils.

wmul_click_utils is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_click_utils is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_click_utils. If not, see <https://www.gnu.org/licenses/>. 
"""
import click
import pytest
from click.testing import CliRunner
from wmul_click_utils import RequiredIf
from wmul_test_utils import make_namedtuple


@pytest.fixture(scope="function")
def setup_required_if():
    @click.command()
    @click.option("--foo", type=str, cls=RequiredIf, required_if="bar", 
                  help="foo help")
    @click.option("--bar", type=str, cls=RequiredIf, required_if="foo", 
                  help="bar help")
    def cli(foo, bar):
        click.echo(f"foo={foo} bar={bar}")

    runner = CliRunner()

    return make_namedtuple(
        "setup_required_if",
        cli=cli,
        runner=runner
    )


def test_both_present(setup_required_if):
    result = setup_required_if.runner.invoke(
        setup_required_if.cli, 
        ["--foo", "foobaz", "--bar", "barbaz"]
    )

    assert not result.exception
    assert result.output == "foo=foobaz bar=barbaz\n"


def test_help(setup_required_if):
    result = setup_required_if.runner.invoke(
        setup_required_if.cli, 
        ["--help"]
    )

    assert "foo help" in result.output
    assert "bar help" in result.output
    

def test_foo_not_present(setup_required_if):
    result = setup_required_if.runner.invoke(
        setup_required_if.cli, 
        ["--bar", "barbaz"]
    )

    assert result.exception
    assert "Illegal usage: foo is required when bar is supplied." in \
        result.output

def test_bar_not_present(setup_required_if):
    result = setup_required_if.runner.invoke(
        setup_required_if.cli, 
        ["--foo", "foobaz"]
    )

    assert result.exception
    assert "Illegal usage: bar is required when foo is supplied." in \
        result.output


def test_neither_present(setup_required_if):
    result = setup_required_if.runner.invoke(
        setup_required_if.cli, 
        []
    )

    assert not result.exception
    assert result.output == "foo=None bar=None\n"


"""
@Author = 'Michael Stanley'

============ Change Log ============
01/17/2023 = Created

============ License ============
Copyright (C) 2023 Michael Stanley

This file is part of wmul_click_utils.

wmul_click_utils is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_click_utils is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_click_utils. If not, see <https://www.gnu.org/licenses/>. 
"""
import click
import pytest
from click.testing import CliRunner
from wmul_click_utils import PAIRED
from wmul_test_utils import make_namedtuple


@pytest.fixture(scope="function")
def setup_paired():
    @click.command()
    @click.option("--foo", type=PAIRED, help="foo help")
    def cli(foo):
        click.echo(", ".join(foo))

    runner = CliRunner()

    return make_namedtuple(
        "setup_paired",
        cli=cli,
        runner=runner
    )


def test_two_given(setup_paired):
    result = setup_paired.runner.invoke(
        setup_paired.cli, 
        ["--foo", "foobaz barbaz"]
    )

    assert not result.exception
    assert "foobaz, barbaz" in result.output


def test_help(setup_paired):
    result = setup_paired.runner.invoke(
        setup_paired.cli, 
        ["--help"]
    )

    assert "foo help" in result.output


def test_only_one_given(setup_paired):
    result = setup_paired.runner.invoke(
        setup_paired.cli, 
        ["--foo", "barbaz"]
    )

    assert result.exception
    assert "barbaz is not a valid Paired Type. The number of space-separated " \
        "values must be even. E.G. '.wav .mp3' and '.wav .mp3 .doc .docx' " \
        "are valid, while '.wav .mp3 .ogg' is not." in result.output


def test_three_given(setup_paired):
    result = setup_paired.runner.invoke(
        setup_paired.cli, 
        ["--foo", "foobaz barbaz bazbaz"]
    )

    assert result.exception
    assert "foobaz barbaz bazbaz is not a valid Paired Type. The number of space-separated " \
        "values must be even. E.G. '.wav .mp3' and '.wav .mp3 .doc .docx' " \
        "are valid, while '.wav .mp3 .ogg' is not." in result.output


def test_four_given(setup_paired):
    result = setup_paired.runner.invoke(
        setup_paired.cli, 
        ["--foo", "foobaz barbaz bazbaz foofoo"]
    )

    assert not result.exception
    assert "foobaz, barbaz, bazbaz, foofoo" in result.output



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wmul_click_utils']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['click']

extras_require = \
{'test': ['pytest', 'wmul_test_utils'], 'tox': ['tox']}

setup(name='wmul_click_utils',
      version='0.1.0',
      description="Additional utilities for use with Python's click cli library.",
      author=None,
      author_email='Michael Stanley <stanley50@marshall.edu>',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



"""
@Author = 'Michael Stanley'

============ Change Log ============
01/17/2023 = Created

============ License ============
Copyright (C) 2023 Michael Stanley

This file is part of wmul_click_utils.

wmul_click_utils is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_click_utils is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_click_utils. If not, see <https://www.gnu.org/licenses/>. 
"""
import click
import pytest
from click.testing import CliRunner
from wmul_click_utils import RequiredUnless
from wmul_test_utils import make_namedtuple


@pytest.fixture(scope="function")
def setup_required_unless():
    @click.command()
    @click.option("--foo", type=str, help="foo help")
    @click.option("--bar", type=str, cls=RequiredUnless, 
        required_unless=["foo"], help="bar help")
    def cli(foo, bar):
        if foo:
            click.echo(f"foo={foo}")
        else:
            click.echo(f"bar={bar}")

    runner = CliRunner()

    return make_namedtuple(
        "setup_required_unless",
        cli=cli,
        runner=runner
    )


def test_both_present(setup_required_unless):
    result = setup_required_unless.runner.invoke(
        setup_required_unless.cli, 
        ["--foo", "foobaz", "--bar", "barbaz"]
    )

    assert not result.exception
    assert "foo=foobaz" in result.output


def test_help(setup_required_unless):
    result = setup_required_unless.runner.invoke(
        setup_required_unless.cli, 
        ["--help"]
    )

    assert "foo help" in result.output
    assert "bar help" in result.output
    

def test_foo_not_present(setup_required_unless):
    result = setup_required_unless.runner.invoke(
        setup_required_unless.cli, 
        ["--bar", "barbaz"]
    )

    assert not result.exception
    assert "bar=barbaz" in result.output


def test_bar_not_present(setup_required_unless):
    result = setup_required_unless.runner.invoke(
        setup_required_unless.cli, 
        ["--foo", "foobaz"]
    )

    assert "foo=foobaz" in result.output
    assert not result.exception


def test_neither_present(setup_required_unless):
    result = setup_required_unless.runner.invoke(
        setup_required_unless.cli, 
        []
    )

    assert result.exception
    assert "Illegal usage: bar is required unless one of 'foo' is supplied."


@pytest.fixture(scope="function")
def setup_required_unless_both_required():
    @click.command()
    @click.option("--foo", type=str, cls=RequiredUnless, 
        required_unless=["bar"], help="foo help")
    @click.option("--bar", type=str, cls=RequiredUnless, 
        required_unless=["foo"], help="bar help")
    def cli(foo, bar):
        if foo:
            click.echo(f"foo={foo}")
        else:
            click.echo(f"bar={bar}")

    runner = CliRunner()

    return make_namedtuple(
        "setup_required_unless_both_required",
        cli=cli,
        runner=runner
    )


def test_both_present_both_required(setup_required_unless_both_required):
    result = setup_required_unless_both_required.runner.invoke(
        setup_required_unless_both_required.cli, 
        ["--foo", "foobaz", "--bar", "barbaz"]
    )

    assert not result.exception
    assert "foo=foobaz" in result.output


def test_help_both_required(setup_required_unless_both_required):
    result = setup_required_unless_both_required.runner.invoke(
        setup_required_unless_both_required.cli, 
        ["--help"]
    )

    assert "foo help" in result.output
    assert "bar help" in result.output
    

def test_foo_not_present_both_required(setup_required_unless_both_required):
    result = setup_required_unless_both_required.runner.invoke(
        setup_required_unless_both_required.cli, 
        ["--bar", "barbaz"]
    )

    assert not result.exception
    assert "bar=barbaz" in result.output


def test_bar_not_present_both_required(setup_required_unless_both_required):
    result = setup_required_unless_both_required.runner.invoke(
        setup_required_unless_both_required.cli, 
        ["--foo", "foobaz"]
    )

    assert "foo=foobaz" in result.output
    assert not result.exception


def test_neither_present_both_required(setup_required_unless_both_required):
    result = setup_required_unless_both_required.runner.invoke(
        setup_required_unless_both_required.cli, 
        []
    )

    assert result.exception
    assert "Illegal usage: bar is required unless one of 'foo' is supplied."


@pytest.fixture(scope="function")
def setup_required_unless_three_required():
    @click.command()
    @click.option("--foo", type=str, cls=RequiredUnless, 
        required_unless=["bar", "baz"], help="foo help")
    @click.option("--bar", type=str, cls=RequiredUnless, 
        required_unless=["foo", "baz"], help="bar help")
    @click.option("--baz", type=str, cls=RequiredUnless, 
        required_unless=["bar", "foo"], help="baz help")
    def cli(foo, bar, baz):
        if foo:
            click.echo(f"foo={foo}")
        elif bar:
            click.echo(f"bar={bar}")
        else:
            click.echo(f"baz={baz}")

    runner = CliRunner()

    return make_namedtuple(
        "setup_required_unless_three_required",
        cli=cli,
        runner=runner
    )


def test_help_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--help"]
    )

    assert "foo help" in result.output
    assert "bar help" in result.output
    assert "baz help" in result.output


# 000
def test_none_present_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        []
    )

    assert result.exception
    assert ("Illegal usage: 'foo' is required unless one of 'baz, bar' is "
        "supplied." in result.output) or ("Illegal usage: 'foo' is required "
        "unless one of 'bar, baz' is supplied." in result.output)


# 001
def test_foo_bar_not_present_three_required(
        setup_required_unless_three_required
    ):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--baz", "bazbaz"]
    )

    assert not result.exception
    assert "baz=bazbaz" in result.output


# 010
def test_foo_baz_not_present_three_required(
        setup_required_unless_three_required
    ):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--bar", "barbaz"]
    )

    assert not result.exception
    assert "bar=barbaz" in result.output


# 011
def test_foo_not_present_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--bar", "barbaz", "--baz", "bazbaz"]
    )

    assert not result.exception
    assert "bar=barbaz" in result.output


# 100
def test_bar_baz_not_present_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--foo", "foobaz"]
    )

    assert "foo=foobaz" in result.output
    assert not result.exception


# 101
def test_bar_not_present_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--foo", "foobaz", "--baz", "bazbaz"]
    )

    assert not result.exception
    assert "foo=foobaz" in result.output


# 110
def test_baz_not_present_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--foo", "foobaz", "--bar", "barbaz"]
    )

    assert not result.exception
    assert "foo=foobaz" in result.output


# 111
def test_all_three_present_three_required(setup_required_unless_three_required):
    result = setup_required_unless_three_required.runner.invoke(
        setup_required_unless_three_required.cli, 
        ["--foo", "foobaz", "--bar", "barbaz", "--baz", "bazbaz"]
    )

    assert not result.exception
    assert "foo=foobaz" in result.output
